"""The DummyEncoder Class is based on Tom Augspurger's great PyData
   Chicago 2016 Talk <https://youtu.be/KLPtEBokqQ0>.
   The original code can be found at git@github.com:TomAugspurger/mtg.git"""
from __future__ import print_function
import numpy as np
import pandas as pd
from sklearn.pipeline import TransformerMixin


class DummyEncoder(TransformerMixin):
    """The DummyEncoder Class converts DataFrame to numpy ndarray (matrix).
       Categorical columns are properly dummy encoded (one hot) to column bit vectors"""

    def __init__(self):
        """Initialize the DummyEncoder class"""
        self.index_ = None
        self.columns_ = None
        self.cat_columns_ = None
        self.non_cat_columns_ = None
        self.cat_map_ = None
        self.cat_blocks_ = None
        self.columns_in_order = None

    def fit_transform(self, df):
        """Fit method for the DummyEncoder"""
        self.index_ = df.index
        self.columns_ = df.columns
        self.cat_columns_ = df.select_dtypes(include=['category']).columns
        self.non_cat_columns_ = df.columns.drop(self.cat_columns_)
        self.cat_map_ = {col: df[col].cat for col in self.cat_columns_}

        # Store all the information about categories/values so we can 'back map' later
        left = len(self.non_cat_columns_)
        self.cat_blocks_ = {}
        for col in self.cat_columns_:
            right = left + len(df[col].cat.categories)
            self.cat_blocks_[col], left = slice(left, right), right

        # This is to ensure that transform always produces the same columns in the same order
        df_with_dummies = pd.get_dummies(df)
        self.columns_in_order = df_with_dummies.columns.tolist()

        # Return the numpy matrix
        return np.asarray(df_with_dummies)

    def transform(self, df):
        """Transform dataframe into a numpy ndarray(matrix)"""

        # Make sure the dataframe columns are the same as the ones passed to fit
        assert df.columns.equals(self.columns_)

        # Convert the dataframe with get_dummies
        df_with_dummies = pd.get_dummies(df)

        # Make sure our columns exactly match what we stored during the fit operation
        my_columns = set(df_with_dummies.columns.tolist())
        output_column_set = set(self.columns_in_order)
        for column in my_columns.difference(output_column_set):
            print('Deleting extra column {:s}'.format(column))
            del df_with_dummies[column]
        for column in output_column_set.difference(my_columns):
            print('Adding column {:s}'.format(column))
            df_with_dummies[column] = 0

        # Returns the same columns in the same order every time
        return np.asarray(df_with_dummies[self.columns_in_order])

    def inverse_transform(self, X):
        """Inverse Transform: numpy ndarray(matrix) to dataframe with original columns"""
        non_cat = pd.DataFrame(X[:, :len(self.non_cat_columns_)],
                               columns=self.non_cat_columns_)
        cats = []
        for col, cat in self.cat_map_.items():
            slice_ = self.cat_blocks_[col]
            codes = X[:, slice_].argmax(1)
            series = pd.Series(pd.Categorical.from_codes(
                codes, cat.categories, ordered=cat.ordered
            ), name=col)
            cats.append(series)
        df = pd.concat([non_cat] + cats, axis=1)[self.columns_]
        return df


# Simple test of the functionality
def test():
    """Test for DummyEncoder Class"""
    import pandas.util.testing as test_utils
    import numpy.testing as numpy_test_utils

    test_df = pd.DataFrame(
        {'A': pd.Categorical(['a', 'b', 'c', 'a'], ordered=True),
         'B': pd.Categorical(['a', 'b', 'c', 'a'], ordered=False),
         'C': pd.Categorical(['a', 'b', 'z', 'a'], categories=['a', 'b', 'z', 'd']),
         'D': [1, 2, 3, 4],
         }
    )
    test_df2 = pd.DataFrame(
        {'A': pd.Categorical(['a', 'b', 'b', 'a'], ordered=True),
         'B': pd.Categorical(['a', 'b', 'd', 'a'], ordered=False),
         'C': pd.Categorical(['a', 'b', 'z', 'y'], categories=['a', 'b', 'z', 'd']),
         'D': [1, 2, 3, 4],
         }
    )

    # Test the transformation from dataframe to numpy ndarray and back again
    encoder = DummyEncoder()
    transformed_data = encoder.fit_transform(test_df)
    back_to_original = encoder.inverse_transform(transformed_data)
    test_utils.assert_frame_equal(back_to_original, test_df)

    # Test that the encoder gives us the same columns on a df with different category values
    transformed_data2 = encoder.transform(test_df2)
    assert transformed_data.shape == transformed_data2.shape

    # First two ROWS should be the same
    numpy_test_utils.assert_equal(transformed_data[0], transformed_data2[0])
    numpy_test_utils.assert_equal(transformed_data[1], transformed_data2[1])


if __name__ == "__main__":
    test()



"""BroMultiLogReader: This class reads in multiple Bro logs.
           Args:
                filepath (str): The full path the file (/full/path/to/the/file.txt) can be a
                                glob (e.g dns*.log) or a gzip file (e.g. dns.log.gz)
"""
from __future__ import print_function
import os
import glob
import gzip
import tempfile
import shutil

# Local Imports
from bat import bro_log_reader


class BroMultiLogReader(object):
    """BroMultiLogReader: This class reads in multiple Bro logs.
           Args:
                filepath (str): The full path the file (/full/path/to/the/file.txt) can be a
                                glob (e.g dns*.log) or a gzip file (e.g. dns.log.gz)
    """

    def __init__(self, filepath):
        """Initialization for the BroMultiLogReader Class"""

        # The filepath may be a glob pattern
        self._files = glob.glob(filepath)

    def readrows(self):
        """The readrows method reads simply 'combines' the rows of
           multiple files OR gunzips the file and then reads the rows
        """

        # For each file (may be just one) create a BroLogReader and use it
        for self._filepath in self._files:

            # Check if the file is zipped
            tmp = None
            if self._filepath.endswith('.gz'):
                tmp = tempfile.NamedTemporaryFile(delete=False)
                with gzip.open(self._filepath, 'rb') as f_in, open(tmp.name, 'wb') as f_out:
                    shutil.copyfileobj(f_in, f_out)

                # Set the file path to the new temp file
                self._filepath = tmp.name

            # Create a BroLogReader
            reader = bro_log_reader.BroLogReader(self._filepath)
            for row in reader.readrows():
                yield row

            # Clean up any temp files
            try:
                if tmp:
                    os.remove(tmp.name)
                    print('Removed temporary file {:s}...'.format(tmp.name))
            except IOError:
                pass


def test():
    """Test for BroMultiLogReader Python Class"""
    from bat.utils import file_utils

    # Grab a test file
    data_path = file_utils.relative_dir(__file__, '../data')

    # For each file, create the Class and test the reader
    files = ['http.log.gz', 'dhcp*.log', 'dhcp*.log.gz']
    for bro_log in files:
        test_path = os.path.join(data_path, bro_log)
        print('Opening Data File: {:s}'.format(test_path))
        reader = BroMultiLogReader(test_path)
        for line in reader.readrows():
            print(line)
    print('Tests successful!')


if __name__ == '__main__':
    # Run the test for easy testing/debugging
    test()



"""LogToDataFrame: Converts a Bro log to a Pandas DataFrame"""
from __future__ import print_function

# Third Party
import pandas as pd

# Local Imports
from bat import bro_log_reader


class LogToDataFrame(pd.DataFrame):
    """LogToDataFrame: Converts a Bro log to a Pandas DataFrame
        Args:
            log_fllename (string): The full path to the Bro log
            ts_index (bool): Set the index to the 'ts' field (default = True)
        Notes:
            This class is fairly simple right now but will probably have additional
            functionality for formal type specifications and performance enhancements
    """
    def __init__(self, log_filename, ts_index=True):
        """Initialize the LogToDataFrame class"""

        # Create a bro reader on a given log file
        reader = bro_log_reader.BroLogReader(log_filename)

        # Create a Pandas dataframe from reader
        super(LogToDataFrame, self).__init__(reader.readrows())

        # Set the index
        if ts_index and not self.empty:
            self.set_index('ts', inplace=True)


# Simple test of the functionality
def test():
    """Test for LogToDataFrame Class"""
    import os
    pd.set_option('display.width', 1000)
    from bat.utils import file_utils

    # Grab a test file
    data_path = file_utils.relative_dir(__file__, '../data')
    test_path = os.path.join(data_path, 'http.log')

    # Convert it to a Pandas DataFrame
    http_df = LogToDataFrame(test_path)

    # Print out the head
    print(http_df.head())

    # Print out the datatypes
    print(http_df.dtypes)

    # Test an empty log (a log with header/close but no data rows)
    test_path = os.path.join(data_path, 'http_empty.log')
    http_df = LogToDataFrame(test_path)

    # Print out the head
    print(http_df.head())

    # Print out the datatypes
    print(http_df.dtypes)

    print('LogToDataFrame Test successful!')


if __name__ == '__main__':
    # Run the test for easy testing/debugging
    test()



"""DataFrameToParquet: Converts a Pandas DataFrame into a Parquet file
   Note:
        Big Thanks to Wes McKinney. This code was borrowed/stolen from
        this article: http://wesmckinney.com/blog/python-parquet-update
"""
from __future__ import print_function

# Third Party
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq


def df_to_parquet(df, filename, compression='SNAPPY'):
    """write_to_parquet: Converts a Pandas DataFrame into a Parquet file
        Args:
            df (pandas dataframe): The Pandas Dataframe to be saved as parquet file
            filename (string): The full path to the filename for the Parquet file
    """

    # Right now there are two open Parquet issues
    # Timestamps in Spark: https://issues.apache.org/jira/browse/ARROW-1499
    # TimeDelta Support: https://issues.apache.org/jira/browse/ARROW-835
    for column in df.columns:
        if(df[column].dtype == 'timedelta64[ns]'):
            print('Converting timedelta column {:s}...'.format(column))
            df[column] = df[column].astype(str)

    arrow_table = pa.Table.from_pandas(df)
    if compression == 'UNCOMPRESSED':
        compression = None
    pq.write_table(arrow_table, filename, compression=compression, use_deprecated_int96_timestamps=True)


def parquet_to_df(filename, use_threads=1):
    """parquet_to_df: Reads a Parquet file into a Pandas DataFrame
        Args:
            filename (string): The full path to the filename for the Parquet file
            ntreads (int): The number of threads to use (defaults to 1)
    """
    try:
        return pq.read_table(filename, use_threads=use_threads).to_pandas()
    except pa.lib.ArrowIOError:
        print('Could not read parquet file {:s}'.format(filename))
        return None


# Simple test of the functionality
def test():
    """Test for methods in this file"""
    import os
    pd.set_option('display.width', 1000)
    from bat.dataframe_to_parquet import df_to_parquet, parquet_to_df
    from bat.log_to_dataframe import LogToDataFrame
    from bat.utils import file_utils
    import tempfile

    # Grab a test file
    data_path = file_utils.relative_dir(__file__, '../data')
    test_path = os.path.join(data_path, 'dns.log')

    # Convert the log to a Pandas DataFrame
    dns_df = LogToDataFrame(test_path)

    # Print out the head
    print(dns_df.head())

    # Create a temporary file
    filename = tempfile.NamedTemporaryFile(delete=False).name

    # Write to a parquet file
    df_to_parquet(dns_df, filename)

    # Read from the parquet file
    new_dns_df = parquet_to_df(filename)

    # Remove temp file
    os.remove(filename)

    # Print out the head
    print(new_dns_df.head())

    # Make sure our conversions didn't lose type info
    assert(dns_df.dtypes.values.tolist() == new_dns_df.dtypes.values.tolist())

    print('DataFrame to Parquet Tests successful!')


if __name__ == '__main__':
    # Run the test for easy testing/debugging
    test()



"""LiveSimulator: This class reads in various Bro logs. The class utilizes
                 the BroLogReader and simply loops over the static bro log
                 file, replaying rows and changing any time stamps
        Args:
            eps (int): Events Per Second that the simulator will emit events (default = 10)
            max_rows (int): The maximum number of rows to generate (default = None (go forever))
"""
from __future__ import print_function
import os
import time
import datetime
import itertools

# Third party
import numpy as np

# Local Imports
from bat import bro_log_reader
from bat.utils import file_utils


class LiveSimulator(object):
    """LiveSimulator: This class reads in various Bro logs. The class utilizes the
                      BroLogReader and simply loops over the static bro log file
                      replaying rows at the specified EPS and changing timestamps to 'now()'
    """

    def __init__(self, filepath, eps=10, max_rows=None):
        """Initialization for the LiveSimulator Class
           Args:
               eps (int): Events Per Second that the simulator will emit events (default = 10)
               max_rows (int): The maximum number of rows to generate (default = None (go forever))
        """

        # Compute EPS timer
        # Logic:
        #     - Normal distribution centered around 1.0/eps
        #     - Make sure never less than 0
        #     - Precompute 1000 deltas and then just cycle around
        self.eps_timer = itertools.cycle([max(0, delta) for delta in np.random.normal(1.0/float(eps), .5/float(eps), size=1000)])

        # Initialize the Bro log reader
        self.log_reader = bro_log_reader.BroLogReader(filepath, tail=False)

        # Store max_rows
        self.max_rows = max_rows

    def readrows(self):
        """Using the BroLogReader this method yields each row of the log file
           replacing timestamps, looping and emitting rows based on EPS rate
        """

        # Loop forever or until max_rows is reached
        num_rows = 0
        while True:

            # Yield the rows from the internal reader
            for row in self.log_reader.readrows():
                yield self.replace_timestamp(row)

                # Sleep and count rows
                time.sleep(next(self.eps_timer))
                num_rows += 1

                # Check for max_rows
                if self.max_rows and (num_rows >= self.max_rows):
                    return

    @staticmethod
    def replace_timestamp(row):
        """Replace the timestamp with now()"""
        if 'ts' in row:
            row['ts'] = datetime.datetime.utcnow()
        return row


def test():
    """Test for LiveSimulator Python Class"""

    # Grab a test file
    data_path = file_utils.relative_dir(__file__, '../data')
    test_path = os.path.join(data_path, 'conn.log')
    print('Opening Data File: {:s}'.format(test_path))

    # Create a LiveSimulator reader
    reader = LiveSimulator(test_path, max_rows=10)
    for line in reader.readrows():
        print(line)
    print('Read with max_rows Test successful!')


if __name__ == '__main__':
    # Run the test for easy testing/debugging
    test()



"""Anomaly Detection Example"""
from __future__ import print_function
import os
import sys
import argparse
import math
from collections import Counter

# Third Party Imports
import pandas as pd
from sklearn.ensemble import IsolationForest
from sklearn.cluster import KMeans

# Local imports
from bat import log_to_dataframe
from bat import dataframe_to_matrix

def entropy(string):
    """Compute entropy on the string"""
    p, lns = Counter(string), float(len(string))
    return -sum(count/lns * math.log(count/lns, 2) for count in p.values())

if __name__ == '__main__':
    # Example to show the dataframe cache functionality on streaming data
    pd.set_option('display.width', 1000)

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('bro_log', type=str, help='Specify a bro log to run BroLogReader test on')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # File may have a tilde in it
    if args.bro_log:
        args.bro_log = os.path.expanduser(args.bro_log)

        # Sanity check either http or dns log
        if 'http' in args.bro_log:
            log_type = 'http'
            features = ['id.resp_p', 'method', 'resp_mime_types', 'request_body_len']
        elif 'dns' in args.bro_log:
            log_type = 'dns'
            features = ['Z', 'rejected', 'proto', 'query', 'qclass_name', 'qtype_name', 'rcode_name', 'query_length', 'answer_length', 'entropy']
        else:
            print('This example only works with Bro with http.log or dns.log files..')
            sys.exit(1)

        # Create a Pandas dataframe from a Bro log
        #bro_df = log_to_dataframe.LogToDataFrame(args.bro_log)
        try:
            bro_df = log_to_dataframe.LogToDataFrame(args.bro_log)
        except IOError:
            print('Could not open or parse the specified logfile: %s' % args.bro_log)
            sys.exit(1)
        print('Read in {:d} Rows...'.format(len(bro_df)))

        # Using Pandas we can easily and efficiently compute additional data metrics
        # Here we use the vectorized operations of Pandas/Numpy to compute query length
        # We'll also compute entropy of the query
        if log_type == 'dns':
            bro_df['query_length'] = bro_df['query'].str.len()
            bro_df['answer_length'] = bro_df['answers'].str.len()
            bro_df['entropy'] = bro_df['query'].map(lambda x: entropy(x))

        # Use the bat DataframeToMatrix class
        to_matrix = dataframe_to_matrix.DataFrameToMatrix()
        bro_matrix = to_matrix.fit_transform(bro_df[features])
        print(bro_matrix.shape)

        # Train/fit and Predict anomalous instances using the Isolation Forest model
        odd_clf = IsolationForest(contamination=0.2) # Marking 20% as odd
        odd_clf.fit(bro_matrix)

        # Now we create a new dataframe using the prediction from our classifier
        odd_df = bro_df[features][odd_clf.predict(bro_matrix) == -1]

        # Now we're going to explore our odd observations with help from KMeans
        odd_matrix = to_matrix.fit_transform(odd_df)
        num_clusters = min(len(odd_df), 4) # 4 clusters unless we have less than 4 observations
        odd_df['cluster'] = KMeans(n_clusters=num_clusters).fit_predict(odd_matrix)
        print(odd_matrix.shape)

        # Now group the dataframe by cluster
        cluster_groups = odd_df[features+['cluster']].groupby('cluster')

        # Now print out the details for each cluster
        print('<<< Outliers Detected! >>>')
        for key, group in cluster_groups:
            print('\nCluster {:d}: {:d} observations'.format(key, len(group)))
            print(group.head())




"""DataFrameToMatrix: Convert a DataFrame to a Numpy Matrix (ndarray) Class"""
from __future__ import print_function

# Third Party
import pandas as pd
import numpy as np

# Local imports
from bat.utils import dummy_encoder


class DataFrameToMatrix(object):
    """DataFrameToMatrix: Convert a DataFrame to a Numpy Matrix (ndarray) Class
        Notes:
            fit_transform: Does a fit and a transform and returns the transformed matrix
            transform: Based on previous fit parameters returns the transformed matrix
    """
    def __init__(self):
        """Initialize the DataFrameToMatrix class"""
        self.column_names = None
        self.cat_columns = None
        self.normalize = True
        self.norm_map = {}
        self.dummy_encoder = dummy_encoder.DummyEncoder()

    def fit_transform(self, input_df, normalize=True):
        """Convert the dataframe to a matrix (numpy ndarray)
        Args:
            input_df (dataframe): The dataframe to convert
            normalize (bool): Boolean flag to normalize numeric columns (default=True)
        """
        # Shallow copy the dataframe (we'll be making changes to some columns)
        _df = input_df.copy(deep=False)

        # Set class variables that will be used both now and later for transform
        self.normalize = normalize

        # Convert columns that are probably categorical
        self.convert_to_categorical(_df)

        # First check for columns that are explicitly categorical
        self.cat_columns = _df.select_dtypes(include=['category']).columns.tolist()

        # Remove any columns that aren't bool/int/float/category
        _df = _df.select_dtypes(include=['bool', 'int', 'float', 'category'])

        # Normalize any numeric columns if normalize specified
        if self.normalize:
            for column in list(_df.select_dtypes(include=[np.number]).columns.values):
                print('Normalizing column {:s}...'.format(column))
                _df[column], _min, _max = self._normalize_series(_df[column])
                self.norm_map[column] = (_min, _max)

        # Now that categorical columns are setup call the dummy_encoder
        return self.dummy_encoder.fit_transform(_df)

    def transform(self, input_df):
        """Convert the dataframe to a matrix (numpy ndarray)
        Args:
            input_df (dataframe): The dataframe to convert
        """

        # Shallow copy the dataframe (we'll be making changes to some columns)
        _df = input_df.copy(deep=False)

        # Convert all columns that are/should be categorical
        for column in self.cat_columns:

            # Sanity check
            if column not in _df:
                raise RuntimeError('Required column {:s} not found'.format(column))

            # If the column isn't already a category then change it
            if _df[column].dtype == 'object':
                print('Changing column {:s} to category'.format(column))
                _df[column] = pd.Categorical(_df[column])

        # Remove any columns that aren't bool/int/float/category
        _df = _df.select_dtypes(include=['bool', 'int', 'float', 'category'])

        # Normalize any numeric columns if normalize specified
        if self.normalize:
            for column in list(_df.select_dtypes(include=[np.number]).columns.values):
                print('Normalizing column {:s}...'.format(column))
                smin, smax = self.norm_map[column]
                _df[column] = (_df[column] - smin) / (smax - smin)

        # Now that categorical columns are setup call the dummy_encoder
        return self.dummy_encoder.transform(_df)

    @staticmethod
    def convert_to_categorical(df):
        """Run a heuristic on the columns of the dataframe to determine whether it contains categorical values
           if the heuristic decides it's categorical then the type of the column is changed
        Args:
            df (dataframe): The dataframe to check for categorical data
        """
        might_be_categorical = df.select_dtypes(include=[object]).columns.tolist()
        for column in might_be_categorical:
            if df[column].nunique() < 20:

                # Convert the column
                print('Changing column {:s} to category...'.format(column))
                df[column] = pd.Categorical(df[column])

    @staticmethod
    def _normalize_series(series):
        smin = series.min()
        smax = series.max()
        if smax - smin == 0:
            print('Cannot normalize series (div by 0) so not normalizing...')
            smin = 0
            smax = 1
        return (series - smin) / (smax - smin), smin, smax


# Simple test of the functionality
def test():
    """Test for DataFrameToMatrix Class"""
    import os
    import pickle
    from tempfile import NamedTemporaryFile
    import numpy.testing as np_test_utils

    test_df = pd.DataFrame(
        {'A': pd.Categorical(['a', 'b', 'c', 'a'], ordered=True),
         'B': pd.Categorical(['a', 'b', 'c', 'a'], ordered=False),
         'C': pd.Categorical(['a', 'b', 'z', 'a'], categories=['a', 'b', 'z', 'd']),
         'D': [1, 2, 3, 4],
         'E': ['w', 'x', 'y', 'z'],
         'F': [1.1, 2.2, 3.3, 4.4],
         'G': pd.to_datetime([0, 1, 2, 3]),
         'H': [True, False, False, True]
         }
    )
    test_df2 = pd.DataFrame(
        {'A': pd.Categorical(['a', 'b', 'b', 'a'], ordered=True),
         'B': pd.Categorical(['a', 'b', 'd', 'a'], ordered=False),
         'C': pd.Categorical(['a', 'b', 'z', 'y'], categories=['a', 'b', 'z', 'd']),
         'D': [1, 2, 3, 7],
         'E': ['w', 'x', 'z', 'foo'],
         'F': [1.1, 2.2, 3.3, 4.4],
         'H': [True, False, False, False]
         }
    )

    # Copy the test_df for testing later
    copy_test_df = test_df.copy()

    # Test the transformation from dataframe to numpy ndarray and back again
    to_matrix = DataFrameToMatrix()
    matrix = to_matrix.fit_transform(test_df)
    matrix_test = to_matrix.transform(test_df)

    # These two matrices should be the same
    np_test_utils.assert_equal(matrix, matrix_test)

    # Assert that the dataframe we passed in didn't change
    copy_test_df.equals(test_df)

    # Test that the conversion gives us the same columns on a df with different category values
    matrix2 = to_matrix.transform(test_df2)
    assert matrix.shape == matrix2.shape

    # First two ROWS should be the same
    np_test_utils.assert_equal(matrix[0], matrix2[0])
    np_test_utils.assert_equal(matrix[1], matrix2[1])

    # Test normalize
    to_matrix_norm = DataFrameToMatrix()
    norm_matrix = to_matrix_norm.fit_transform(test_df)
    print(norm_matrix)
    assert(norm_matrix[:, 0].min() == 0)
    assert(norm_matrix[:, 0].max() == 1)

    # Make sure normalize 'does the right thing' when doing transform
    norm_matrix2 = to_matrix_norm.transform(test_df2)
    assert(norm_matrix2[:, 0].min() == 0)
    assert(norm_matrix2[:, 0].max() == 2)    # Normalization is based on FIT range

    # Test div by zero in normalize
    test_df3 = test_df2.copy()
    test_df3['D'] = [1, 1, 1, 1]
    norm_matrix3 = to_matrix_norm.fit_transform(test_df3)
    assert(norm_matrix3[:, 0].min() == 1)
    assert(norm_matrix3[:, 0].max() == 1)

    # Test serialization
    temp = NamedTemporaryFile(delete=False)
    pickle.dump(to_matrix, temp)
    temp.close()

    # Deserialize and test
    to_matrix_from_disk = pickle.load(open(temp.name, 'rb'))
    matrix3 = to_matrix_from_disk.transform(test_df)
    matrix4 = to_matrix_from_disk.transform(test_df2)
    np_test_utils.assert_equal(matrix, matrix3)
    np_test_utils.assert_equal(matrix2, matrix4)

    # Remove the temporary file
    os.unlink(temp.name)


if __name__ == "__main__":
    test()



"""Bro log to Parquet Dataframe Example"""
from __future__ import print_function
import os
import sys
import argparse

# Local imports
from bat.log_to_parquet import log_to_parquet

if __name__ == '__main__':
    # Example to write Parquet file from a bro log

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('bro_log', type=str, help='Specify a bro log to run BroLogReader test on')
    parser.add_argument('-o', '--parquet-file', type=str, required=True, help='Specify the parquet file to write to')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # File may have a tilde in it
    if args.bro_log and args.parquet_file:
        args.bro_log = os.path.expanduser(args.bro_log)
        args.parquet_file = os.path.expanduser(args.parquet_file)

        # Write out the parquet file
        print('Writing Parquet file: {:s}'.format(args.parquet_file))
        log_to_parquet(args.bro_log, args.parquet_file)



"""Bro log to Pandas Dataframe Example"""
from __future__ import print_function
import os
import sys
import argparse

# Local imports
from bat.log_to_dataframe import LogToDataFrame

if __name__ == '__main__':
    # Example to populate a Pandas dataframe from a bro log reader

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('bro_log', type=str, help='Specify a bro log to run BroLogReader test on')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # File may have a tilde in it
    if args.bro_log:
        args.bro_log = os.path.expanduser(args.bro_log)

        # Create a Pandas dataframe from a Bro log
        bro_df = LogToDataFrame(args.bro_log)

        # Print out the head of the dataframe
        print(bro_df.head())

        # Print out the types of the columns
        print(bro_df.dtypes)



"""Signal Catching utility"""
from __future__ import print_function

import signal
import sys
import time
from contextlib import contextmanager


@contextmanager
def signal_catcher(callback):
    """Catch signals and invoke the callback method"""
    def _catch_exit_signal(sig_num, _frame):
        print('Received signal {:d} invoking callback...'.format(sig_num))
        callback()

    signal.signal(signal.SIGINT, _catch_exit_signal)
    signal.signal(signal.SIGQUIT, _catch_exit_signal)
    signal.signal(signal.SIGTERM, _catch_exit_signal)
    yield


def my_exit():
    """My exit callback (just for testing)"""
    print('My Exit got called...Do some cleanup or whatever...')
    sys.exit()


def test():
    """Test the SignalCatcher"""
    with signal_catcher(my_exit):
        time.sleep(1)


if __name__ == '__main__':
    test()



"""Risky DNS bat Example"""
from __future__ import print_function
import os
import sys
import argparse
from pprint import pprint
import pickle

# Third Party Imports
try:
    import tldextract
except ImportError:
    print('\nThis example needs tldextract. Please do a $pip install tldextract and rerun this example')
    sys.exit(1)

# Local imports
from bat import bro_log_reader
from bat.utils import vt_query, signal_utils

def save_vtq():
    """Exit on Signal"""
    global vtq

    print('Saving VirusTotal Query Cache...')
    pickle.dump(vtq, open('vtq.pkl', 'wb'), protocol=pickle.HIGHEST_PROTOCOL)
    sys.exit()

if __name__ == '__main__':
    # Risky DNS/VT Query application
    global vtq

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('bro_log', type=str, help='Specify a bro log to run BroLogReader test on')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # Sanity check that this is a dns log
    if 'dns' not in args.bro_log:
        print('This example only works with Bro dns.log files..')
        sys.exit(1)

    # File may have a tilde in it
    if args.bro_log:
        args.bro_log = os.path.expanduser(args.bro_log)

        # See if we have a serialized VirusTotal Query Class.
        # If we do not have one we'll create a new one
        try:
            vtq = pickle.load(open('vtq.pkl', 'rb'))
            print('Opening VirusTotal Query Cache (cache_size={:d})...'.format(vtq.size))
        except IOError:
            vtq = vt_query.VTQuery(max_cache_time=60*24*7) # One week cache

        # See our 'Risky Domains' Notebook for the analysis and
        # statistical methods used to compute this risky set of TLDs
        risky_tlds = set(['info', 'tk', 'xyz', 'online', 'club', 'ru', 'website', 'in', 'ws',
                          'top', 'site', 'work', 'biz', 'name', 'tech', 'loan', 'win', 'pro'])

        # Launch long lived process with signal catcher
        with signal_utils.signal_catcher(save_vtq):

            # Run the bro reader on the dns.log file looking for risky TLDs
            reader = bro_log_reader.BroLogReader(args.bro_log)
            for row in reader.readrows():

                # Pull out the TLD
                query = row['query']
                tld = tldextract.extract(query).suffix

                # Check if the TLD is in the risky group
                if tld in risky_tlds:
                    # Make the query with the full query
                    results = vtq.query_url(query)
                    if results.get('positives', 0) > 3: # At least four hits
                        print('\nRisky Domain DNS Query Found')
                        print('From: {:s} To: {:s} QType: {:s} RCode: {:s}'.format(row['id.orig_h'],
                               row['id.resp_h'], row['qtype_name'], row['rcode_name']))
                        pprint(results)

        # Save the Virus Total Query
        save_vtq()



"""Run a VirusTotal Query on Extracted File Hashes"""
from __future__ import print_function
import os
import sys
import argparse
from pprint import pprint

# Local imports
from bat import bro_log_reader
from bat.utils import vt_query

if __name__ == '__main__':
    """Run a VirusTotal Query on Extracted File Hashes"""

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('bro_log', type=str, help='Specify a bro log to run BroLogReader test on')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # Sanity check that this is a file log
    if 'files' not in args.bro_log:
        print('This example only works with Bro files.log files..')
        sys.exit(1)

    # File may have a tilde in it
    if args.bro_log:
        args.bro_log = os.path.expanduser(args.bro_log)

        # Create a VirusTotal Query Class
        vtq = vt_query.VTQuery()

        # Run the bro reader on a given log file
        reader = bro_log_reader.BroLogReader(args.bro_log, tail=True)
        for row in reader.readrows():
            file_sha = row.get('sha256', '-') # Bro uses - for empty field
            if file_sha == '-':
                file_sha = row.get('sha1', '-') # Bro uses - for empthy field
                if file_sha == '-':
                    print('Should not find a sha256 or a sha1 key! Skipping...')
                    continue

            # Make the query with either sha
            results = vtq.query_file(file_sha)
            if results.get('positives', 0) > 1: # At least two hits
                pprint(results)




"""FileTailer Python Class"""
from __future__ import print_function
import os
import sys
import argparse
import time
import math
from collections import Counter
import json

# Third Party Imports
import pandas as pd
from sklearn.ensemble import IsolationForest
from sklearn.cluster import KMeans
from sklearn.cluster import MiniBatchKMeans
import hdbscan

# Local imports
from bat import bro_log_reader, live_simulator
from bat import dataframe_to_matrix, dataframe_cache


if __name__ == '__main__':
    # Example to show the dataframe cache functionality on streaming data
    pd.set_option('display.width', 200)

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('bro_log', type=str, help='Specify a bro log to run BroLogReader test on')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # File may have a tilde in it
    if args.bro_log:
        args.bro_log = os.path.expanduser(args.bro_log)


        # Sanity check for either http or dns log
        if 'http' in args.bro_log:
            log_type = 'http'
            features = ['id.resp_p', 'method', 'resp_mime_types', 'request_body_len']
        elif 'dns' in args.bro_log:
            log_type = 'dns'
            features = ['Z', 'rejected', 'proto', 'query', 'qclass_name', 'qtype_name', 'rcode_name', 'query_length']
        else:
            print('This example only works with Bro with http.log or dns.log files..')
            sys.exit(1)

        # Create a Bro log reader
        print('Opening Data File: {:s}'.format(args.bro_log))
        reader = bro_log_reader.BroLogReader(args.bro_log, tail=True)

        # Create a Bro IDS log live simulator
        print('Opening Data File: {:s}'.format(args.bro_log))
        reader = live_simulator.LiveSimulator(args.bro_log, eps=10) # 10 events per second

        # Create a Dataframe Cache
        df_cache = dataframe_cache.DataFrameCache(max_cache_time=600)  # 10 minute cache

        # Streaming Clustering Class
        batch_kmeans = MiniBatchKMeans(n_clusters=4, verbose=True)
        # num_clusters = min(len(odd_df), 4) # 4 clusters unless we have less than 4 observations

        # Use the BroThon DataframeToMatrix class
        features = ['Z', 'rejected', 'proto', 'query', 'qclass_name', 'qtype_name', 'rcode_name', 'query_length', 'id.resp_p']
        to_matrix = dataframe_to_matrix.DataFrameToMatrix()

        # Add each new row into the cache
        time_delta = 10
        timer = time.time() + time_delta
        FIRST_TIME = True
        for row in reader.readrows():
            df_cache.add_row(row)

            # Every 30 seconds grab the dataframe from the cache
            if time.time() > timer:
                timer = time.time() + time_delta

                # Get the windowed dataframe (10 minute window)
                bro_df = df_cache.dataframe()

                # Add query length
                bro_df['query_length'] = bro_df['query'].str.len()

                # Use the bat DataframeToMatrix class
                features = ['Z', 'rejected', 'proto', 'query', 'qclass_name', 'qtype_name', 'rcode_name', 'query_length', 'id.resp_p']
                to_matrix = dataframe_to_matrix.DataFrameToMatrix()
                bro_matrix = to_matrix.fit_transform(bro_df[features])
                print(bro_matrix.shape)

                # Print out the range of the daterange and some stats
                print('DataFrame TimeRange: {:s} --> {:s}'.format(str(bro_df['ts'].min()), str(bro_df['ts'].max())))

                # Train/fit and Predict anomalous instances using the Isolation Forest model
                odd_clf = IsolationForest(contamination=0.2) # Marking 20% as odd
                odd_clf.fit(bro_matrix)
                bro_df['anomalous'] = [predict==-1 for predict in odd_clf.predict(bro_matrix)]

                # Now we create a new dataframe using the prediction from our classifier
                odd_df = bro_df[bro_df['anomalous']]

                # Now we're going to explore our odd observations with help from KMeans
                odd_matrix = to_matrix.transform(odd_df[features])
                #clusters = KMeans(n_clusters=num_clusters).fit_predict(odd_matrix).tolist()
                batch_kmeans.partial_fit(odd_matrix)
                clusters = batch_kmeans.predict(odd_matrix).tolist()

                # Set the cluster number for all the entries in the original dataframe
                bro_df['cluster'] = [-1 if not anom else clusters.pop(0) for anom in bro_df['anomalous'] ]

                # Now group the dataframe by cluster
                cluster_groups = bro_df.groupby('cluster')

                # Now print out the details for each cluster
                show_fields = ['id.orig_h', 'id.resp_h'] + features
                print('<<< Outliers Detected! >>>')
                for key, group in cluster_groups:
                    print('\nCluster {:d}: {:d} observations'.format(key, len(group)))
                    print(group[show_fields].head())

                # Output to file
                with open('streaming_output.json', 'w') as fp:
                    for row in bro_df.to_dict('records'):
                        row['ts'] = str(row['ts'])
                        #print(row)
                        json.dump(row, fp)
                        fp.write('\n')



"""VTQuery Class"""

from __future__ import print_function
import collections
import time
import pprint

# Third Party
import requests

# Local imports
from bat.utils import cache


class VTQuery(object):
    """VTQuery: Will query VirusTotal for a sha, the class contains a Cache to avoid
                slamming duplicate queries.

        Args:
            apikey (str): The API key to use for VirusTotal queries (default=None)
            summary (bool): Just return summary information for VTQuery (default=True)
            max_cache_size (int): Maximum size of query cache (default=1000)
            max_cache_time (int): Time to keep query results in cache (default=60 minutes)
    """

    def __init__(self, apikey=None, summary=True, max_cache_size=1000, max_cache_time=60, throttle=True):
        """VTQuery Init"""

        # Public VT API Key
        # Note: The Virus Total key below is a low-volume public key.
        #       Please call this method with your own VT API key :)
        if apikey is None:
            print('Using public VT API Key: Please set apikey=<your key> when creating this class')
        pub_vt_apikey = 'ab0933e5b4d8032031bbce54b4170453e62c229dcf93fb99b0b80f09e415f809'
        self.apikey = apikey or pub_vt_apikey
        self.exclude = ['scan_id', 'md5', 'sha1', 'sha256', 'resource', 'response_code', 'permalink',
                        'verbose_msg', 'scans'] if summary else []
        self.throttle = throttle

        # Create query cache
        self.query_cache = cache.Cache(max_size=max_cache_size, timeout=max_cache_time*60)  # Convert to Seconds

    def query_file(self, file_sha, verbose=False):
        """Query the VirusTotal Service
            Args:
               file_sha (str): The file sha1 or sha256 hash
               url (str): The domain/url to be queried (default=None)
        """

        # Sanity check sha hash input
        if len(file_sha) not in [64, 40]:  # sha256 and sha1 lengths
            print('File sha looks malformed: {:s}'.format(file_sha))
            return {'file_sha': file_sha, 'malformed': True}

        # Call and return the internal query method
        return self._query('file', file_sha, verbose)

    def query_url(self, url, verbose=False):
        """Query the VirusTotal Service
            Args:
               url (str): The domain/url to be queried
        """
        # Call and return the internal query method
        return self._query('url', url, verbose)

    @property
    def size(self):
        return self.query_cache.size

    def _query(self, query_type, query_str, verbose=False):
        """Internal query method for the VirusTotal Service
            Args:
               query_type(str): The type of query (either 'file' or 'url')
               query_str (str): The file hash or domain/url to be queried
        """
        # First check query cache
        cached = self.query_cache.get(query_str)
        if cached:
            if verbose:
                print('Returning Cached VT Query Results')
            return cached

        # Not in cache so make the actual query
        if query_type == 'file':
            response = requests.get('https://www.virustotal.com/vtapi/v2/file/report',
                                    params={'apikey': self.apikey, 'resource': query_str, 'allinfo': 1})
        else:
            response = requests.post('https://www.virustotal.com/vtapi/v2/url/report',
                                     params={'apikey': self.apikey, 'resource': query_str, 'allinfo': 1})

        # Make sure we got a json blob back
        try:
            vt_output = response.json()
        except ValueError:
            error_msg = 'VirusTotal no valid response, throttling and trying again...'
            if self.throttle:
                if verbose:
                    print(error_msg)
                time.sleep(30)
                return self._query(query_type, query_str)

            return {'vt_error': error_msg}

        # Check for not-found
        if not vt_output or vt_output['response_code'] == 0:
            output = {'query': query_str, 'not_found': True}
            self.query_cache.set(query_str, output)
            return output

        # Exclude some fields (if summary=True)
        output = {field: vt_output[field] for field in vt_output.keys() if field not in self.exclude}

        # Put the file sha in the output
        output['query'] = query_str

        # Organize the scans fields
        scan_results = collections.Counter()
        for scan in vt_output['scans'].values():
            if 'result' in scan:
                if scan['result']:
                    scan_results[scan['result']] += 1
        output['scan_results'] = scan_results.most_common(5)

        # Pull results in Cache
        self.query_cache.set(query_str, output)

        # Return results
        return output


# Unit test: Create the class and test it
def test():
    """vt_query.py test"""

    # Execute the worker (unit test)
    vt_query = VTQuery(summary=False)
    output = vt_query.query_file('eb107c004e6e1bbd3b32ad7961661bbe28a577b0cb5dac4cfd518f786029cb95')
    print('\n<<< Unit Test Full>>>')
    pprint.pprint(output)
    vt_query = VTQuery()
    output = vt_query.query_file('4ecf79302ba0439f62e15d0526a297975e6bb32ea25c8c70a608916a609e5a9c')
    print('\n<<< Unit Test Summary>>>')
    pprint.pprint(output)

    # Test queries on domain names
    output = vt_query.query_url('amazon.co.uk.security-check.ga')
    print('\n<<< Unit Test Domain Names>>>')
    pprint.pprint(output)

    # Test Cache
    output = vt_query.query_file('4ecf79302ba0439f62e15d0526a297975e6bb32ea25c8c70a608916a609e5a9c')
    print('\n<<< Unit Test Cache>>>')
    pprint.pprint(output)

    # Test Size
    assert vt_query.size == 2

    # Test some error conditions
    output = vt_query.query_file('123')
    print('\n<<< Unit Test Malformed SHA HASH>>>')
    pprint.pprint(output)
    output = vt_query.query_file('123f79302ba0439f62e15d0526a297975e6bb32ea25c8c70a608916a609e5a9c')
    print('\n<<< Unit Test Not Found>>>')
    pprint.pprint(output)


if __name__ == "__main__":
    test()



"""File utilities that might be useful"""
from __future__ import print_function
import os


def all_files_in_directory(path):
    """Recursively list all files under a directory

        Args:
            path: the path of the directory to traverse
        Returns:
            a list of all the files contained withint the directory
    """
    file_list = []
    for dirname, _dirnames, filenames in os.walk(path):
        for filename in filenames:
            # Skip OS Files
            if filename != '.DS_Store':
                file_list.append(os.path.join(dirname, filename))
    return file_list


def most_recent(path, startswith=None, endswith=None):
    """Recursively inspect all files under a directory and return the most recent

        Args:
            path (str): the path of the directory to traverse
            startswith (str): the file name start with (optional)
            endswith (str): the file name ends with (optional)
        Returns:
            the most recent file within the subdirectory
    """
    candidate_files = []
    for filename in all_files_in_directory(path):
        if startswith and not os.path.basename(filename).startswith(startswith):
            continue
        if endswith and not filename.endswith(endswith):
            continue
        candidate_files.append({'name': filename, 'modtime': os.path.getmtime(filename)})

    # Return the most recent modtime
    most_recent = sorted(candidate_files, key=lambda k: k['modtime'], reverse=True)
    return most_recent[0]['name'] if most_recent else None


def file_dir(file_path):
    """Root directory for a file_path

        Args:
            file_path: a fully qualified file path
        Returns:
            the directory which contains the file
    """
    return os.path.dirname(os.path.realpath(file_path))


def relative_dir(file_path, rel_dir):
    """Relative directory to the file_path

        Args:
            file_path: a fully qualified file path
        Returns:
            the relative directory
    """
    return os.path.join(file_dir(file_path), rel_dir)


def test_utils():
    """Test the utility methods"""

    print('Relative Dir: {:s}'.format(relative_dir(__file__, '.')))
    print('File Directory: {:s}'.format(file_dir(__file__)))
    path = file_dir(__file__)
    print('Path: {:s}'.format(path))
    for my_file in all_files_in_directory(path):
        print('\t%s' % my_file)

    print('Most Recent: {:s}'.format(most_recent(path)))
    print('Most Recent Python File: {:s}'.format(most_recent(path, endswith='.py')))

    # Test startswith
    assert most_recent(path, startswith='cache', endswith='.py') == relative_dir(__file__, 'cache.py')

    # Test when no filename match
    assert most_recent(path, endswith='.nomatch') is None


if __name__ == '__main__':
    test_utils()



"""DirWatcher: Watches a directory and calls the callback when files are created"""

import os
import time

# Other Imports
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

# Local Imports
from bat.utils import file_utils


class DirWatcher(FileSystemEventHandler):
    """Watches a directory and calls the callback when files are created/modified"""

    def __init__(self, data_dir, callback, **kwargs):
        """Initialization"""
        self.callback = callback
        self.kwargs = kwargs

        # Now setup dynamic monitoring of the data directory
        observer = Observer()
        observer.schedule(self, path=data_dir)
        observer.start()

    def on_any_event(self, event):
        """File created or modified"""
        if os.path.isfile(event.src_path):
            self.callback(event.src_path, **self.kwargs)


def my_callback(file_path):
    """Callback for new file"""
    print('New File Created: {:s}'.format(file_path))


def test():
    """Test the DirWatcher Class"""
    watch_path = file_utils.relative_dir(__file__, '../../data')
    print('Watching Directory: %s' % watch_path)
    DirWatcher(watch_path, my_callback)

    # Create a file and then delete it
    temp_file = os.path.join(watch_path, 'test.tmp')
    open(temp_file, 'w').close()
    time.sleep(1)
    os.remove(temp_file)


if __name__ == '__main__':
    test()



"""DataFrameToParquet: Converts a Pandas DataFrame into a Parquet file
   Note:
        Big Thanks to Wes McKinney. This code was borrowed/stolen from
        this article: http://wesmckinney.com/blog/python-parquet-update
"""
from __future__ import print_function

# Third Party
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

# Local Imports
from bat.bro_log_reader import BroLogReader


def _make_df(rows):
    """Internal Method to make and clean the dataframe in preparation for sending to Parquet"""

    # Make DataFrame
    df = pd.DataFrame(rows).set_index('ts')

    # TimeDelta Support: https://issues.apache.org/jira/browse/ARROW-835
    for column in df.columns:
        if(df[column].dtype == 'timedelta64[ns]'):
            print('Converting timedelta column {:s}...'.format(column))
            df[column] = df[column].astype(str)
    return df


def log_to_parquet(bro_log, parquet_file, compression='SNAPPY', row_group_size=1000000):
    """write_to_parquet: Converts a Bro log into a Parquet file
        Args:
            bro_log (string: The full path to the bro log to be saved as a Parquet file
            parquet_file (string): The full path to the filename for the Parquet file
            compression (string): The compression algo to use (defaults to 'SNAPPY')
            row_group_size (int): The size of the parquet row groups (defaults to 1000000)
        Notes:
            Right now there are two open Parquet issues
            - Timestamps in Spark: https://issues.apache.org/jira/browse/ARROW-1499
            - TimeDelta Support: https://issues.apache.org/jira/browse/ARROW-835
    """

    # Set up various parameters
    current_row_set = []
    writer = None
    num_rows = 0

    # Spin up the bro reader on a given log file
    reader = BroLogReader(bro_log)
    for num_rows, row in enumerate(reader.readrows()):

        # Append the row to the row set
        current_row_set.append(row)

        # If we have enough rows add to the Parquet table
        if num_rows % row_group_size == 0:
            print('Writing {:d} rows...'.format(num_rows))
            if writer is None:
                arrow_table = pa.Table.from_pandas(_make_df(current_row_set))
                writer = pq.ParquetWriter(parquet_file, arrow_table.schema, compression=compression, use_deprecated_int96_timestamps=True)
                writer.write_table(arrow_table)
            else:
                arrow_table = pa.Table.from_pandas(_make_df(current_row_set))
                writer.write_table(arrow_table)

            # Empty the current row set
            current_row_set = []

    # Add any left over rows and close the Parquet file
    if num_rows:
        print('Writing {:d} rows...'.format(num_rows))
        arrow_table = pa.Table.from_pandas(_make_df(current_row_set))
        writer.write_table(arrow_table)
        writer.close()
        print('Parquet File Complete')


# Simple test of the functionality
def test():
    """Test for methods in this file"""
    import os
    pd.set_option('display.width', 1000)
    from bat.dataframe_to_parquet import parquet_to_df
    from bat.log_to_dataframe import LogToDataFrame
    from bat.utils import file_utils
    import tempfile

    # Grab a test file
    data_path = file_utils.relative_dir(__file__, '../data')
    test_path = os.path.join(data_path, 'dns.log')

    # Convert the log to a Pandas DataFrame
    dns_df = LogToDataFrame(test_path)

    # Print out the head
    print(dns_df.head())

    # Create a temporary file
    filename = tempfile.NamedTemporaryFile(delete=False).name

    # Write to a parquet file
    log_to_parquet(test_path, filename)

    # Read from the parquet file
    new_dns_df = parquet_to_df(filename)

    # Remove temp file
    os.remove(filename)

    # Print out the head
    print(new_dns_df.head())

    # Make sure our conversions didn't lose type info
    # TODO: Uncomment this test when the following PR is fixed
    #       - TimeDelta Support: https://issues.apache.org/jira/browse/ARROW-835
    # assert(dns_df.dtypes.values.tolist() == new_dns_df.dtypes.values.tolist())

    # Test an empty log (a log with header/close but no data rows)
    test_path = os.path.join(data_path, 'http_empty.log')
    filename = tempfile.NamedTemporaryFile(delete=False).name
    log_to_parquet(test_path, filename)
    parquet_to_df(filename)
    os.remove(filename)

    print('DataFrame to Parquet Tests successful!')


if __name__ == '__main__':
    # Run the test for easy testing/debugging
    test()



"""Network utilities that might be useful"""
from __future__ import print_function

import socket
import binascii

# Local imports
from bat.utils import compat


def mac_to_str(address):
    """Convert a MAC address to a readable/printable string

       Args:
           address (str): a MAC address in hex form (e.g. '\x01\x02\x03\x04\x05\x06')
       Returns:
           str: Printable/readable MAC address
    """
    return ':'.join('%02x' % compat.ord(b) for b in address)


def str_to_mac(mac_string):
    """Convert a readable string to a MAC address

           Args:
               mac_string (str): a readable string (e.g. '01:02:03:04:05:06')
           Returns:
               str: a MAC address in hex form
        """
    sp = mac_string.split(':')
    mac_string = ''.join(sp)
    return binascii.unhexlify(mac_string)


def inet_to_str(inet):
    """Convert inet object to a string

        Args:
            inet (inet struct): inet network address
        Returns:
            str: Printable/readable IP address
    """
    # First try ipv4 and then ipv6
    try:
        return socket.inet_ntop(socket.AF_INET, inet)
    except ValueError:
        return socket.inet_ntop(socket.AF_INET6, inet)


def str_to_inet(address):
    """Convert an a string IP address to a inet struct

        Args:
            address (str): String representation of address
        Returns:
            inet: Inet network address
    """
    # First try ipv4 and then ipv6
    try:
        return socket.inet_pton(socket.AF_INET, address)
    except socket.error:
        return socket.inet_pton(socket.AF_INET6, address)


def is_internal(ip_address):
    """Determine if the address is an internal ip address
       Note: This is super bad, improve it
    """
    # Local networks 10.0.0.0/8, 172.16.0.0/12, '192.168.0.0/16
    local_nets = '10.', '172.16.', '192.168.', '169.254', 'fd', 'fe80::'
    return any([ip_address.startswith(local) for local in local_nets])


def is_special(ip_address):
    """Determine if the address is SPECIAL
       Note: This is super bad, improve it
    """
    special = {'224.0.0.251': 'multicast_dns',
               'ff02::fb': 'multicast_dns'}
    return special[ip_address] if ip_address in special else False


def test_utils():
    """Test the utility methods"""

    print(mac_to_str(b'\x01\x02\x03\x04\x05\x06'))
    assert mac_to_str(b'\x01\x02\x03\x04\x05\x06') == '01:02:03:04:05:06'
    assert str_to_mac('01:02:03:04:05:06') == b'\x01\x02\x03\x04\x05\x06'
    foo = b'\x01\x02\x03\x04\x05\x06'
    bar = mac_to_str(foo)
    assert str_to_mac(bar) == foo
    print(inet_to_str(b'\x91\xfe\xa0\xed'))
    assert inet_to_str(b'\x91\xfe\xa0\xed') == '145.254.160.237'
    assert str_to_inet('145.254.160.237') == b'\x91\xfe\xa0\xed'
    assert is_internal('10.0.0.1')
    assert is_internal('222.2.2.2') is False
    assert is_special('224.0.0.251')
    assert is_special('224.0.0.252') is False
    print('Success!')


if __name__ == '__main__':
    test_utils()



"""Silly data generator (Faker (https://github.com/joke2k/faker) and others
   are much better, but we just need something simple"""
from __future__ import print_function
import string

# Third Party
import pandas as pd
import numpy as np


def df_random(num_numeric=3, num_categorical=3, num_rows=100):
    """Generate a dataframe with random data. This is a general method
       to easily generate a random dataframe, for more control of the
       random 'distributions' use the column methods (df_numeric_column, df_categorical_column)
       For other distributions you can use numpy methods directly (see example at bottom of this file)
        Args:
            num_numeric (int): The number of numeric columns (default = 3)
            num_categorical (int): The number of categorical columns (default = 3)
            num_rows (int): The number of rows to generate  (default = 100)
    """

    # Construct DataFrame
    df = pd.DataFrame()
    column_names = string.ascii_lowercase

    # Create numeric columns
    for name in column_names[:num_numeric]:
        df[name] = df_numeric_column(num_rows=num_rows)

    # Create categorical columns
    for name in column_names[num_numeric:num_numeric+num_categorical]:
        df[name] = df_categorical_column(['foo', 'bar', 'baz'], num_rows=num_rows)

    # Return the dataframe
    return df


def df_numeric_column(min_value=0, max_value=1, num_rows=100):
    """Generate a numeric column with random data
        Args:
            min_value (float): Minimum value (default = 0)
            max_value (float): Maximum value (default = 1)
            num_rows (int): The number of rows to generate  (default = 100)
    """
    # Generate numeric column
    return pd.Series(np.random.uniform(min_value, max_value, num_rows))


def df_categorical_column(category_values, num_rows=100, probabilities=None):
    """Generate a categorical column with random data
        Args:
            category_values (list): A list of category values (e.g. ['red', 'blue', 'green'])
            num_rows (int): The number of rows to generate  (default = 100)
            probabilities (list): A list of probabilities of each value (e.g. [0.6, 0.2, 0.2]) (default=None  an equal probability)
    """
    splitter = np.random.choice(range(len(category_values)), num_rows, p=probabilities)
    return pd.Series(pd.Categorical.from_codes(splitter, categories=category_values))


def test():
    """Test the data generator methods"""
    df = df_random()
    print('Random DataFrame')
    print(df.head())

    # Test the numerical column generator
    df['delta_v'] = df_numeric_column(-100, 100)
    print('\nNumerical column generator (added delta_v)')
    print(df.head())

    # Test the categorical column generator
    df['color'] = df_categorical_column(['red', 'green', 'blue'])
    print('\nCategorical column generator (added color)')
    print(df.head())

    # Test the categorical column generator with probabilities
    df['color'] = df_categorical_column(['red', 'green', 'blue'], probabilities=[0.6, 0.3, 0.1])
    print('\nProbabilities should be ~60% red, %30 green and %10 blue')
    print(df['color'].value_counts())

    # Also we can just use the built in Numpy method for detailed control
    # over the numeric distribution
    my_series = pd.Series(np.random.normal(0, 1, 1000))
    print('\nStats on numpy normal (gaussian) distribution')
    print(my_series.describe())


if __name__ == '__main__':
    test()



"""BroLogReader: This class reads in various Bro logs. The class inherits from
                 the FileTailer class so it supports the following use cases:
                   - Read contents of a Bro log file        (tail=False)
                   - Read contents + 'tail -f' Bro log file (tail=True)
       Args:
            filepath (str): The full path the file (/full/path/to/the/file.txt)
            delimiter (str): The delimiter in the Bro logs (default='\t')
            tail (bool): Do a dynamic tail on the file (i.e. tail -f) (default=False)
"""
from __future__ import print_function
import os
import time
import datetime

# Local Imports
from bat.utils import file_tailer, file_utils


class BroLogReader(file_tailer.FileTailer):
    """BroLogReader: This class reads in various Bro logs. The class inherits from
                     the FileTailer class so it supports the following use cases:
                       - Read contents of a Bro log file        (tail=False)
                       - Read contents + 'tail -f' Bro log file (tail=True)
           Args:
                filepath (str): The full path the file (/full/path/to/the/file.txt)
                delimiter (str): The delimiter in the Bro logs (default='\t')
                tail (bool): Do a dynamic tail on the file (i.e. tail -f) (default=False)
                strict (bool): Raise an exception on conversions errors (default=False)
    """

    def __init__(self, filepath, delimiter='\t', tail=False, strict=False):
        """Initialization for the BroLogReader Class"""

        # First check if the file exists and is readable
        if not os.access(filepath, os.R_OK):
            raise IOError('Could not read/access bro log file: {:s}'.format(filepath))

        # Setup some class instance vars
        self._filepath = filepath
        self._delimiter = delimiter
        self._tail = tail
        self._strict = strict

        # Setup the Bro to Python Type mapper
        self.field_names = []
        self.field_types = []
        self.type_converters = []
        self.type_mapper = {'bool': lambda x: True if x == 'T' else False,
                            'count': int,
                            'int': int,
                            'double': float,
                            'time': lambda x: datetime.datetime.fromtimestamp(float(x)),
                            'interval': lambda x: datetime.timedelta(seconds=float(x)),
                            'string': lambda x: x,
                            'enum': lambda x: x,
                            'port': int,
                            'unknown': lambda x: x}
        self.dash_mapper = {'bool': False, 'count': 0, 'int': 0, 'port': 0, 'double': 0.0,
                            'time': datetime.datetime.fromtimestamp(86400), 'interval': datetime.timedelta(seconds=0),
                            'string': '-', 'unknown:': '-'}

        # Initialize the Parent Class
        super(BroLogReader, self).__init__(self._filepath, full_read=True, tail=self._tail)

    def readrows(self):
        """The readrows method reads in the header of the Bro log and
            then uses the parent class to yield each row of the log file
            as a dictionary of {key:value, ...} based on Bro header.
        """
        # Calling the internal _readrows so we can catch issues/log rotations
        reconnecting = True
        while True:
            # Yield the rows from the internal reader
            try:
                for row in self._readrows():
                    if reconnecting:
                        print('Successfully monitoring {:s}...'.format(self._filepath))
                        reconnecting = False
                    yield row
            except IOError:
                # If the tail option is set then we do a retry (might just be a log rotation)
                if self._tail:
                    print('Could not open file {:s} Retrying...'.format(self._filepath))
                    reconnecting = True
                    time.sleep(5)
                    continue
                else:
                    break

            # If the tail option is set then we do a retry (might just be a log rotation)
            if self._tail:
                print('File closed {:s} Retrying...'.format(self._filepath))
                reconnecting = True
                time.sleep(5)
                continue
            else:
                break

    def _readrows(self):
        """Internal method _readrows, see readrows() for description"""

        # Read in the Bro Headers
        offset, self.field_names, self.field_types, self.type_converters = self._parse_bro_header(self._filepath)

        # Use parent class to yield each row as a dictionary
        for line in self.readlines(offset=offset):

            # Check for #close
            if line.startswith('#close'):
                return

            # Yield the line as a dict
            yield self.make_dict(line.strip().split(self._delimiter))

    def _parse_bro_header(self, bro_log):
        """Parse the Bro log header section.

            Format example:
                #separator \x09
                #set_separator	,
                #empty_field	(empty)
                #unset_field	-
                #path	httpheader_recon
                #fields	ts	origin	useragent	header_events_json
                #types	time	string	string	string
        """

        # Open the Bro logfile
        with open(bro_log, 'r') as bro_file:

            # Skip until you find the #fields line
            _line = bro_file.readline()
            while not _line.startswith('#fields'):
                _line = bro_file.readline()

            # Read in the field names
            field_names = _line.strip().split(self._delimiter)[1:]

            # Read in the types
            _line = bro_file.readline()
            field_types = _line.strip().split(self._delimiter)[1:]

            # Setup the type converters
            type_converters = []
            for field_type in field_types:
                type_converters.append(self.type_mapper.get(field_type, self.type_mapper['unknown']))

            # Keep the header offset
            offset = bro_file.tell()

        # Return the header info
        return offset, field_names, field_types, type_converters

    def make_dict(self, field_values):
        ''' Internal method that makes sure any dictionary elements
            are properly cast into the correct types.
        '''
        data_dict = {}
        for key, value, field_type, converter in zip(self.field_names, field_values, self.field_types, self.type_converters):
            try:
                # We have to deal with the '-' based on the field_type
                data_dict[key] = self.dash_mapper.get(field_type, '-') if value == '-' else converter(value)
            except ValueError as exc:
                print('Conversion Issue for key:{:s} value:{:s}\n{:s}'.format(key, str(value), str(exc)))
                data_dict[key] = value
                if self._strict:
                    raise exc

        return data_dict


def test():
    """Test for BroLogReader Python Class"""
    import pytest

    # Grab a test file
    data_path = file_utils.relative_dir(__file__, '../data')

    # For each file, create the Class and test the reader
    files = ['app_stats.log', 'conn.log', 'dhcp.log', 'dns.log', 'files.log', 'ftp.log',
             'http.log', 'notice.log', 'smtp.log', 'ssl.log', 'weird.log', 'x509.log']
    for bro_log in files:
        test_path = os.path.join(data_path, bro_log)
        print('Opening Data File: {:s}'.format(test_path))
        reader = BroLogReader(test_path, tail=False)  # First with no tailing
        for line in reader.readrows():
            print(line)
    print('Read with NoTail Test successful!')

    # Test an empty log (a log with header/close but no data rows)
    test_path = os.path.join(data_path, 'http_empty.log')
    reader = BroLogReader(test_path)
    for line in reader.readrows():
        print(line)

    # Test some of the error conditions
    reader.field_names = ['good', 'error']
    reader.type_converters = [int, lambda x: datetime.datetime.fromtimestamp(float(x))]
    reader.make_dict([5, '0, .5, .5'])

    # Test invalid file path
    with pytest.raises(IOError):
        BroLogReader('nowhere.log')

    # Now include tailing (note: as an automated test this needs to timeout quickly)
    try:
        from interruptingcow import timeout

        # Spin up the class
        tailer = BroLogReader(test_path, tail=True)

        # Tail the file for 2 seconds and then quit
        try:
            with timeout(2, exception=RuntimeError):
                for line in tailer.readrows():
                    print(line)
        except RuntimeError:  # InterruptingCow raises a RuntimeError on timeout
            print('Tailing Test successful!')

    except ImportError:
        print('Tailing Test not run, need interruptcow module...')


if __name__ == '__main__':
    # Run the test for easy testing/debugging
    test()



"""FileTailer Python Class"""
from __future__ import print_function
import os
import sys
import argparse
from pprint import pprint

# Local imports
from bat import bro_log_reader

if __name__ == '__main__':
    # Example to run the bro log reader on a given file

    # Collect args from the command line
    parser = argparse.ArgumentParser()
    parser.add_argument('bro_log', type=str, help='Specify a bro log to run BroLogReader test on')
    parser.add_argument('-t', '--tail', action='store_true', help='Turn on log tailing')
    args, commands = parser.parse_known_args()

    # Check for unknown args
    if commands:
        print('Unrecognized args: %s' % commands)
        sys.exit(1)

    # File may have a tilde in it
    if args.bro_log:
        args.bro_log = os.path.expanduser(args.bro_log)

        # Run the bro reader on a given log file
        reader = bro_log_reader.BroLogReader(args.bro_log, tail=args.tail, strict=True)
        for row in reader.readrows():
            pprint(row)



"""Plotting utilities"""


def plot_defaults():
    try:
        import matplotlib.pyplot as plt
        plt.style.use('seaborn-muted')
        plt.rc('font', size=14)
        plt.rc('xtick', labelsize=12)
        plt.rc('ytick', labelsize=12)
        plt.rc('axes', labelsize=14)
        plt.rc('axes', axisbelow=True)
        plt.rc('grid', color='grey')
        plt.rc('grid', alpha=.5)
        plt.rc('grid', linestyle='-')
        plt.rcParams['figure.figsize'] = 16.0, 6.0
        # plt.rcParams.update({'figure.autolayout': True})
        plt.rc('patch', force_edgecolor=True)
    except ImportError:
        print('Could not import matplotlib... this is fine...')


def test():
    """Test the Plot Utilities"""
    plot_defaults()


if __name__ == '__main__':
    test()



""" DataFrame Cache Class: The class provides a caching mechanism for a Pandas DataFrame """
from __future__ import print_function
import time
from collections import deque
import pandas as pd


class DataFrameCache(object):
    """DataFrameCache: The class provides a caching mechanism for a Pandas DataFrame.

        Args:
            max_cache_size (int): Maximum size of dataframe cache (default=10000)
            max_cache_time (int): Time to keep rows in cache (default=60 seconds)
    """
    def __init__(self, max_cache_size=100000, max_cache_time=60):
        """Initialize the DataFrameCache class"""
        self.max_size = max_cache_size
        self.max_time = max_cache_time
        self.row_deque = deque(maxlen=self.max_size)
        self.time_deque = deque(maxlen=self.max_size)
        self._dataframe = pd.DataFrame()

    def add_row(self, row):
        """Add a row to the DataFrameCache class"""
        self.add_rows([row])

    def add_rows(self, list_of_rows):
        """Add a list of rows to the DataFrameCache class"""
        for row in list_of_rows:
            self.row_deque.append(row)
            self.time_deque.append(time.time())
        # Update the data structure
        self.update()

    def dataframe(self):
        """Return a DataFrame with the current window of data
           Note: Only call this when you want the dataframe to be reconstructed"""
        self.update()
        return pd.DataFrame(list(self.row_deque))

    def update(self):
        """Update the deque, removing rows based on time"""
        expire_time = time.time() - self.max_time
        while self.row_deque and self.time_deque[0] < expire_time:
            self.row_deque.popleft()  # FIFO
            self.time_deque.popleft()


# Simple test of the functionality
def test():
    """Test for DataFrameCache Class"""
    import copy

    df_cache = DataFrameCache(max_cache_size=10, max_cache_time=1)  # Make it small and short for testing

    # Make some fake data
    base_row = {'id': 0, 'foo': 'bar', 'port': 80, 'protocol': 17}

    # Create an array of test rows
    test_data = []
    for i in range(20):
        row = copy.deepcopy(base_row)
        row['id'] = i
        test_data.append(row)

    # Add rows
    df_cache.add_rows(test_data)

    # Make sure the cache size is working properly
    my_df = df_cache.dataframe()
    assert len(my_df) == 10
    assert my_df.iloc[0]['id'] == 10  # This means the deque is proper FIFO

    # Now test time expiration
    time.sleep(1)
    my_df = df_cache.dataframe()
    assert len(my_df) == 0


if __name__ == "__main__":
    test()
